import os
import urllib.parse
import re
import shutil
import ffmpeg
import heic2png


class Attachment:
    """ Class for holding information about an attachment """
    def __init__(self, database, rowid: str, filename: str, mime_type: str,
                 copy=False, copy_directory=None, home_directory=os.environ['HOME']) -> None:
        """
                        Parameters
                        ----------
                        database : imessagedb.DB
                            An instance of a connected database

                        rowid, filename, mime_type : str
                            Fields from the database

                        copy : bool
                            Whether or not to copy the attachment, default is False

                        copy_directory : str
                            The directory to copy the attachment to, default is none

                        home_directory : str
                            Needed to know where to get the attachments from. The attachments in the database
                            are listed under ~/Library/Messages/Attachments, so need to be able to translate that

                        """
        self._database = database
        self._rowid = rowid
        self._filename = filename
        self._mime_type = mime_type
        self._copy = copy
        self._copy_directory = copy_directory
        self._home_directory = home_directory

        self._destination_path = None
        self._popup_type = None
        self._conversion_type = None
        self._skip = False
        self._missing = False
        self._needs_conversion = False
        self._force = self._database.control.getboolean('force copy', False)

        # The path is set to use ~, so replace it with the home directory
        self._original_path = self._filename.replace('~', self._home_directory)
        if not os.path.exists(self._original_path):
            self._missing = True
            return

        if self._copy_directory is None or not os.path.isdir(self._copy_directory):
            self._copy = False

        if self._copy:
            parts = self._original_path.split('/')
            last = parts.pop()
            penultimate = parts.pop()
            self._destination_filename = f'{penultimate}-{last}'

        self._process_mime_type()  # We may need to do a conversion and therefore change the filename
        if self._copy:
            self._destination_path = f'{self._copy_directory}/{self._destination_filename}'
            if len(self._destination_path) > 200: # Some filenames are too long
                self._destination_filename = f'{self._destination_filename[:50]}---{self._destination_filename[-50:]}'
                self._destination_path = f'{self._copy_directory}/{self._destination_filename}'
        else:
            self._destination_path = self._original_path

        return

    @property
    def rowid(self) -> str:
        """ Return the rowid """
        return self._rowid

    @property
    def filename(self) -> str:
        """ Return the filename of the attachment """
        return self._filename

    @property
    def mime_type(self) -> str:
        """ Return the mime_type of the attachment"""
        return self._mime_type

    @property
    def copy(self) -> bool:
        """ Return if the attachment should be copied """
        return self._copy

    @property
    def destination_path(self) -> str:
        """ Return the destination path the attachment will be copied to """
        return self._destination_path

    @property
    def popup_type(self) -> str:
        """ Return the popup type, one of [Audio, Video, Picture] """
        return self._popup_type

    @property
    def conversion_type(self) -> str:
        """ Return the conversion type, one of [Audio, HEIC, Video] """
        return self._conversion_type

    @property
    def skip(self) -> bool:
        """ Return if we need to skip this attachment """
        return self._skip

    @property
    def missing(self) -> bool:
        """ Return if the attachment is missing """
        return self._missing

    @property
    def original_path(self) -> str:
        """ Return the original path of the attachment """
        return self._original_path

    @property
    def destination_filename(self) -> str:
        """ Return the filename at the destination (not the whole path)"""
        return self._destination_filename

    @property
    def html_path(self) -> str:
        """ Return the html escaped path """
        return urllib.parse.quote(self.destination_path)

    @property
    def link_path(self) -> str:
        """ Return a link to the attachment """
        return f"file://{urllib.parse.quote(self.destination_path)}"

    def _process_mime_type(self) -> None:
        """ Based on the mime_type, decide how to convert the file if necessary and how to rename it """
        if self._mime_type is None:
            # The only type of file that doesn't have a mime type that I am currently interested in are audio files
            search_name = re.search(".caf$", self._filename)
            if search_name:
                self._popup_type = "Audio"
                self._conversion_type = "Audio"
                if self.copy:
                    self._destination_filename = f'{self._destination_filename}.mp3'
            else:
                self._skip = True  # We don't care about this attachment

        elif self._mime_type[0:5] == "image":
            self._popup_type = "Picture"
            if self.mime_type == 'image/heic':
                self._conversion_type = "HEIC"
                if self._copy:
                    self._destination_filename = f'{self._destination_filename}.png'

        # Process audio
        elif self._mime_type[0:5] == "audio":
            self._popup_type = "Audio"
            self._conversion_type = "Audio"
            if self._copy:
                self._destination_filename = f'{self._destination_filename}.mp3'

        # Process Video
        elif self._mime_type[0:5] == "video":
            self._popup_type = "Video"
            if self.mime_type != "video/mp4":
                self._conversion_type = "Video"
                if self.copy:
                    self._destination_filename = f'{self._destination_filename}.mp4'

    def copy_attachment(self) -> None:
        """ Copy the attachment """
        # Skip the file copy if the copy already exists
        if self._force or not os.path.exists(self._destination_path):
            print(f"Copying {self._destination_filename}")
            try:
                shutil.copyfile(self._original_path, self._destination_path)
                return
            except Exception as exp:
                print(f"Failed to copy {self._destination_filename}: {exp}")
                return
        else:
            # If the file already exists, do nothing
            return

    def convert_heic_image(self, heic_location, png_location) -> None:
        """ Convert a HEIC image to a PNG so it can be viewed in the browser """
        # Don't do the expensive conversion if we've already converted it
        if self._force or not os.path.exists(png_location):
            try:
                print(f"Converting {os.path.basename(png_location)}")
                heic_image = heic2png.HEIC2PNG(heic_location)
                heic_image.save(png_location)
                return
            except Exception as exp:
                print(f'Failed to convert {heic_location} to {png_location}: {exp}')
                return
        else:
            # If the file exists already, don't convert it
            return

    def convert_audio_video(self, original, converted) -> None:
        """ Convert an audio or video file from an undisplayable source to something the browser can display"""
        if self._force or not os.path.exists(converted):
            try:
                print(f"Converting {os.path.basename(converted)}")
                stream = ffmpeg.input(original)
                stream = ffmpeg.output(stream, converted)
                stream = ffmpeg.overwrite_output(stream)
                ffmpeg.run(stream, quiet=True)
                return
            except Exception as exp:
                print(f'Failed to convert {original} to {converted}: {exp}')
                return
        else:
            # If the file exists already, don't convert it
            return



import plistlib
import datetime

from . import attachments
import imessagedb


def _date_from_webkit(webkit_timestamp: float) -> datetime.datetime:
    # From https://www.epochconverter.com/webkit
    epoch_start = datetime.datetime(1601,1,1)
    delta = datetime.timedelta(microseconds=int(webkit_timestamp*1000000))
    return epoch_start + delta


def _convert_attributed_body(encoded: bytes) -> str:
    # This general logic was copied from
    # https://github.com/my-other-github-account/imessage_tools/blob/master/imessage_tools.py, however
    # I needed to make some improvements

    text = encoded.split(b'NSNumber')[0]
    text = text.split(b'NSString')[1]
    text = text.split(b'NSDictionary')[0]
    text = text[6:-12]
    if b'\x01' in text:
        text = text.split(b'\x01')[1]
    if b'\x00' in text:
        text = text.split(b'\x00')[1]
    if b'\x86' in text:
        text = text.split(b'\x86')[0]
    return text.decode('utf-8', errors='replace')


class Message:
    """ Class for holding information about a message """

    def __init__(self, database, rowid, guid, date, is_from_me, handle_id, attributed_body, message_summary_info, text,
                 reply_to_guid, thread_originator_guid, thread_originator_part, chat_id, message_attachments):
        """
                Parameters
                ----------
                database : imessagedb.DB
                    An instance of a connected database

                rowid, guid, date, is_from_me, handle_id, attributed_body, message_summary_info, text,
                 reply_to_guid, thread_originator_guid, thread_originator_part, chat_id : str
                    The parameters are the fields in the database

                message_attachments : list
                    The attachments for this message"""

        self._rowid = rowid
        self._guid = guid
        self._date = date
        self._is_from_me = is_from_me
        self._handle_id = handle_id
        self._attributed_body = attributed_body
        self._message_summary_info = message_summary_info
        self._text = text
        self._attributed_body = attributed_body
        self._reply_to_guid = reply_to_guid
        self._thread_originator_guid = thread_originator_guid
        self._thread_originator_part = thread_originator_part
        self._chat_id = chat_id
        self._attachments = message_attachments
        self._thread = {}
        self._edits = []

        # There are a lot of messages that are saved into attributed_body instead of the text field.
        #  There isn't a good way to convert this in Python that I've found, so I have to run a
        #  program to do it. I need to fix this.

        if (self._text is None or text == '' or text == ' ') and self._attributed_body is not None:
            self._text = _convert_attributed_body(self._attributed_body)

        # Edits are stored in message_summary_info
        try:
            plist = plistlib.loads(self._message_summary_info)
            if 'ec' in plist:
                for row in plist['ec']['0']:
                    self._edits.append(_convert_attributed_body(row['t']))
                    # row['d'] = _date_from_webkit(row['d'])
        except plistlib.InvalidFileException as exp:
            pass

    def __repr__(self) -> str:
        return_string = f'RowID: {self._rowid}' \
                        f' GUID: {self._guid}' \
                        f' Date: {self._date}' \
                        f' From me: {self._is_from_me}' \
                        f' HandleID: {self._handle_id}' \
                        f' Message: {self._text}' \
                        f' Originator Thread: {self._thread_originator_guid}' \
                        f' Reply Message: {self._reply_to_guid}' \
                        f' Thread Part: {self._thread_originator_part}' \
                        f' Chat ID: {self._chat_id}' \
                        f' Attachments: {self._attachments}'
        return str(return_string)

    @property
    def rowid(self) -> str:
        return self._rowid

    @property
    def guid(self) -> str:
        return self._guid

    @property
    def date(self) -> datetime.datetime:
        return self._date

    @property
    def is_from_me(self) -> bool:
        return self._is_from_me

    @property
    def handle_id(self) -> str:
        return self._handle_id

    @property
    def attributed_body(self) -> bytes:
        return self._attributed_body

    @property
    def message_summary_info(self) -> bytes:
        return self._message_summary_info

    @property
    def text(self) -> str:
        return self._text

    @property
    def edits(self) -> list:
        return self._edits

    @property
    def reply_to_guid(self) -> str:
        return self._reply_to_guid

    @property
    def thread_originator_guid(self) -> str:
        return self._thread_originator_guid

    @property
    def chat_id(self) -> str:
        return self._chat_id

    @property
    def attachments(self) -> attachments.Attachments:
        return self._attachments

    @property
    def thread(self) -> dict:
        return self._thread



from termcolor import colored
from datetime import datetime
import imessagedb


def _print_thread(thread_header, current_message: str) -> str:
    thread_string = ""
    thread_list = thread_header.thread
    thread_list[thread_header.rowid] = thread_header
    for i in sorted(thread_list.values(), key=lambda x: x.date):
        if i == current_message:
            break
        attachment_string = ""
        if i.attachments is not None:
            attachment_string = f" Attachments: {i.attachments}"
        thread_string = f'{thread_string}[{i.text}{attachment_string}] '
    return thread_string


class TextOutput:
    """ Creates a text file (or string) from a Messages list

        ...

        There are a number of options in the configuration file that affect how the HTML is created.
        In the DISPLAY section, the following impact the output:

         me = Me :
                    The name to put for your part of the conversation. It defaults to 'Me'.

        use text color = True :
                    If False, don't color the text at all

        me text color = blue
        them text color = magenta :
                    The color for the names in text output. The color can only be one of the following options:
                    black, red, green, yellow, blue, magenta, cyan, white, light_grey, dark_grey,
                    light_red, light_green, light_yellow, light_blue, light_magenta, light_cyan

        reply text color = light_grey :
                    The color for the reply text """

    def __init__(self, database, me: str, person: str, messages, output_file=None) -> None:
        self._database = database
        self._me = me
        self._person = person
        self._messages = messages
        self._attachment_list = self._database.attachment_list
        self._output_file = output_file

        self._use_color = self._database.config['DISPLAY'].getboolean('use text color', fallback=True)
        self._me_color = self._database.config['DISPLAY'].get('me text color', fallback="blue")
        self._them_color = self._database.config['DISPLAY'].get('them text color', fallback="magenta")
        self._reply_color = self._database.config['DISPLAY'].get('reply text color', fallback="light_grey")

        self._string_array = []
        self._get_messages()
        return

    def _get_messages(self) -> None:
        for message in self._messages.message_list:
            date = message.date

            day = datetime(int(date[0:4]), int(date[5:7]), int(date[8:10]),
                           int(date[11:13]), int(date[14:16]),
                           int(date[17:19])).strftime('%a')

            if message.is_from_me:
                who = self._color(self._me, self._me_color, attrs=['bold'])
            else:
                who = self._color(self._person, self._them_color, attrs=['bold'])

            reply_to = ""
            attachment_string = ""

            if message.attachments:
                attachment_string = f'Attachments: {message.attachments}'
            if message.thread_originator_guid:
                if message.thread_originator_guid in self._messages.guids:
                    original_message = self._messages.guids[message.thread_originator_guid]
                    reply_to = self._color(f'Reply to: {_print_thread(original_message, message)}',
                                           self._reply_color)
            self._string_array.append(f'<{day} {date}> {who}: {message.text} {reply_to} {attachment_string}')

    def _color(self, text: str, color: str, attrs=None) -> str:
        if self._use_color:
            if attrs:
                return colored(text, color, attrs=attrs)
            else:
                return colored(text, color)
        else:
            return text

    def save(self) -> None:
        """ Save the text output to the file """
        print('\n'.join(self._string_array), file=self._output_file)
        return

    def print(self) -> None:
        """ Print the text output to stdout """
        print('\n'.join(self._string_array))
        return

    def __repr__(self) -> str:
        return '\n'.join(self._string_array)



from imessagedb.attachment import Attachment
from alive_progress import alive_bar


class Attachments:
    """ All attachments """
    def __init__(self, database, copy=None, copy_directory=None) -> None:
        """
            Parameters
            ----------
            database : imessagedb.DB
                An instance of a connected database

            copy : bool
                Whether or not to copy attachments, default is to use the configuration parameter

            copy_directory : str
                The directory to copy attachments into
        """

        self._database = database
        if copy is None:
            self._copy = self._database.control.getboolean('copy', fallback=False)
        else:
            self._copy = copy

        if copy_directory is None:
            self._copy_directory = self._database.control.get('attachment directory', fallback="/tmp/attachments")
        else:
            self._copy_directory = copy_directory

        self._attachment_list = {}
        self._message_join = {}

        # Get the list of all the attachments

        self._database.connection.execute('select count(rowid)from attachment')
        (row_count_total) = self._database.connection.fetchone()
        row_count_total = row_count_total[0]

        self._database.connection.execute('select rowid, filename, mime_type from attachment')

        i = self._database.connection.fetchone()
        with alive_bar(row_count_total, title="Getting Attachments", stats="({rate}, eta: {eta})") as bar:
            while i:
                rowid = i[0]
                filename = i[1]
                mime_type = i[2]
                if filename is not None:
                    self.attachment_list[rowid] = Attachment(self._database, rowid, filename, mime_type,
                                                             copy=self._copy,
                                                             copy_directory=self._copy_directory)
                bar()
                i = self._database.connection.fetchone()

        # Get the join of attachments and messages

        self._database.connection.execute('select message_id, attachment_id from message_attachment_join')
        i = self._database.connection.fetchone()
        while i:
            message_id = i[0]
            attachment_id = i[1]
            if message_id in self.message_join:
                self.message_join[message_id].append(attachment_id)
            else:
                self.message_join[message_id] = [attachment_id]
            i = self._database.connection.fetchone()

        return

    @property
    def attachment_list(self) -> dict:
        """ Return the list of all attachments """
        return self._attachment_list

    @property
    def message_join(self) -> dict:
        """ Return the mapping of messages to attachments """
        return self._message_join

    def __iter__(self):
        return self.attachment_list.__iter__()

    def __len__(self) -> int:
        return len(self.attachment_list.keys())



from imessagedb.chat import Chat


class Chats:
    """ All handles in the database """

    def __init__(self, database) -> None:
        """
            Parameters
            ----------
            database : imessagedb.DB
                An instance of a connected database"""

        self._database = database
        self._chat_list = {}
        self._chat_identifiers = {}

        self._get_chats()
        return

    def __repr__(self) -> str:
        string_array = []
        for i in sorted(self.chat_list):
            string_array.append(str(self.chat_list[i]))
        return '\n'.join(string_array)

    def __len__(self) -> int:
        return len(self._chat_list)

    def _get_chats(self):
        self._database.connection.execute('select rowid, chat_identifier, display_name from chat')
        rows = self._database.connection.fetchall()
        for row in rows:
            rowid = row[0]
            chat_identifier = row[1]
            display_name = row[2]
            new_chat = Chat(self._database, rowid, chat_identifier, display_name)
            self.chat_list[new_chat.rowid] = new_chat
            if new_chat.chat_identifier in self.chat_identifiers:
                self.chat_identifiers[new_chat.chat_identifier].append(new_chat)
            else:
                self.chat_identifiers[new_chat.chat_identifier] = [new_chat]

        # Add the last chat date to all the chats
        for i in self.chat_list.values():
            select_string = "select " \
                    "datetime(max(message_date)/1000000000 + strftime('%s', '2001-01-01'),'unixepoch','localtime') " \
                    f"from chat_message_join cmj where chat_id = {i.rowid}"

            self._database.connection.execute(select_string)
            rows = self._database.connection.fetchall()
            i.last_message_date = rows[0][0]

        # Add the participants for all the chats
        self._database.connection.execute('select chat_id, handle_id from chat_handle_join')
        rows = self._database.connection.fetchall()
        for row in rows:
            chat_id = row[0]
            handle_id = row[1]

            self.chat_list[row[0]].add_participant(row[1])

        return

    @property
    def chat_list(self) -> dict:
        """ Return the list of chats by rowid in a dict"""
        return self._chat_list

    @property
    def chat_identifiers(self) -> dict:
        """ Return the list of chats by chat identifier in a dict"""
        return self._chat_identifiers



from imessagedb.handle import Handle


class Handles:
    """ All handles in the database """

    def __init__(self, database) -> None:
        """
            Parameters
            ----------
            database : imessagedb.DB
                An instance of a connected database
        """
        self._database = database
        self._handle_list = {}
        self._numbers = {}

        self._get_handles()
        return

    def _get_handles(self):
        self._database.connection.execute('select rowid, id, service from handle')
        rows = self._database.connection.fetchall()
        for row in rows:
            rowid = row[0]
            number = row[1]
            service = row[2]
            new_handle = Handle(self._database, rowid, number, service)
            self._handle_list[new_handle.rowid] = new_handle
            if new_handle.number in self._numbers:
                self._numbers[new_handle.number].append(new_handle)
            else:
                self._numbers[new_handle.number] = [new_handle]

    @property
    def handles(self) -> dict:
        """ Return the list of handles """
        return self._handle_list

    @property
    def numbers(self) -> dict:
        """ Return the list of handles indexed by the number """
        return self._numbers

    def __iter__(self):
        return self._handle_list

    def __len__(self) -> int:
        return len(self._handle_list)

    def __repr__(self) -> str:
        handle_array = []
        for i in sorted(self._handle_list.keys()):
            handle_array.append(self._handle_list[i])
        return '\n'.join(map(str, handle_array))



"""
Provides access to the iMessage chat.db, including parsing and output

"""
from importlib.metadata import version

# read version from installed package

__version__ = version("imessagedb")

import os
import configparser
import logging
import argparse
import sys
from datetime import datetime
from imessagedb.db import DB


DEFAULT_CONFIGURATION = '''
[CONTROL]

# Whether or not to copy the attachments into a different directory. This is needed for two reasons:
#  1) The web browser does not have access to the directory that the attachments are stored, so it cannot display them
#  2) Some of the attachments need to be converted in order to be viewed in the browser

copy = True

# The directory to put the output. The html file will go in {copy_directory}/{Person}.html,
#   and the attachments will go in {copy_directory}/{Person}_Attachments

copy directory = /tmp

# If the file already exists in the destination directory it is not recopied, but that can be overridden by
#  specifying 'force copy' as true

force copy = False

# 'Skip attachments' ignores attachments

skip attachments = False

# Additional details show some other information on each message, including the chat id and any edits that have 
#  been done

additional details = False

# Some extra verbosity if true

verbose = True

[DISPLAY]

# Output type, either html or text

output_type = html

# Inline attachments mean that the images are in the HTML instead of loaded when hovered over

inline attachments = False

# Popup location is where the attachment popup window shows up, and is either 'upper right', 'upper left' or 'floating'

popup location = upper right

# 'me' is the name to put for your text messages
me = Me

# The color for the name in text output. No color is used if 'use text color' is false.
#  The color can only be one of the following options:
#   black, red, green, yellow, blue, magenta, cyan, white, light_grey, dark_grey,
#   light_red, light_green, light_yellow, light_blue, light_magenta, light_cyan

use text color = True
me text color = blue
them text color = magenta
reply text color = light_grey

# The background color of the text messages for you and for the other person in html output
# The options for colors can be found here: https://www.w3schools.com/cssref/css_colors.php

me html background color = AliceBlue
them html background color = Lavender

# The background color of the thread in replies

thread background = HoneyDew
me thread background = AliceBlue
them thread background = Lavender

# The color of the name in the html output

me html name color = Blue
them html name color = Purple


[CONTACTS]

# A person that you text with can have multiple numbers, and you may not always want to specify the full specific
#  number as stored in the handle database, so you can do the mapping here, providing the name of a person,
#  and a comma separated list of numbers

Samantha: +18434676040, samanthasmilt@gmail.com, 
   s12ddd2@colt.edu
Abe: +16103499696
Marissa: +14029490739
'''


def _create_default_configuration(filename: str) -> None:
    """Generates a default configuration if one is not passed in

    """

    f = open(filename, "w")
    f.write(DEFAULT_CONFIGURATION)
    f.close()
    return


def get_contacts(configuration: configparser.ConfigParser) -> dict:
    result = {}
    contact_list = configuration.items('CONTACTS')
    for contact in contact_list:
        key = contact[0]
        value = contact[1]

        # Get rid of spaces and newlines
        stripped = value.replace('\n', '').replace(' ', '')
        number_list = stripped.split(',')
        result[key] = number_list
    return result


def run() -> None:
    logging.basicConfig(level=logging.INFO, format='%(asctime)s <%(name)s> %(message)s', datefmt='%Y-%m-%d %H:%M:%S')
    logger = logging.getLogger('main')
    logger.debug("Processing parameters")

    argument_parser = argparse.ArgumentParser()
    person_mutex_group = argument_parser.add_mutually_exclusive_group()
    person_mutex_group.add_argument('--handle', help="A list of handles to search against", nargs='*')
    person_mutex_group.add_argument("--name", help="Person to get conversations about")
    argument_parser.add_argument("-c", "--configfile", help="Location of the configuration file",
                                 default=f'{os.environ["HOME"]}/.config/iMessageDB.ini')
    argument_parser.add_argument("-o", "--output_directory",
                                 help="The output directory where the output and attachments go")
    argument_parser.add_argument("--database", help="The database file to open",
                                 default=f"{os.environ['HOME']}/Library/Messages/chat.db")
    argument_parser.add_argument("-m", "--me", help="The name to use to refer to you", default="Me")
    argument_parser.add_argument("-t", "--output_type", help="The type of output", choices=["text", "html"])
    argument_parser.add_argument("-i", "--inline", help="Show the attachments inline", action="store_true")
    copy_mutex_group = argument_parser.add_mutually_exclusive_group()
    copy_mutex_group.add_argument("-f", "--force", help="Force a copy of the attachments", action="store_true")
    copy_mutex_group.add_argument("--no_copy", help="Don't copy the attachments", action="store_true")
    argument_parser.add_argument("--no_attachments", help="Don't process attachments at all", action="store_true")
    argument_parser.add_argument("-v", "--verbose", help="Turn on additional output", action="store_true")
    argument_parser.add_argument('--start_time', help="The start time of the messages in YYYY-MM-DD HH:MM:SS format")
    argument_parser.add_argument('--end_time', help="The end time of the messages in YYYY-MM-DD HH:MM:SS format")
    argument_parser.add_argument('--version', help="Prints the version number", action="store_true")

    args = argument_parser.parse_args()

    if args.version:
        print(f"imessagedb {__version__}", file=sys.stderr)
        exit(0)

    # First read in the configuration file, creating it if need be, then overwrite the values from the command line
    if not os.path.exists(args.configfile):
        _create_default_configuration(args.configfile)
    config = configparser.ConfigParser()
    config.read(args.configfile)

    CONTROL = 'CONTROL'
    DISPLAY = 'DISPLAY'

    config.set(CONTROL, 'verbose', str(args.verbose))

    if not args.name and not args.handle:
        argument_parser.print_help(sys.stderr)
        print("\n ** You must supply either a name or one or more handles")
        exit(1)

    person = None
    numbers = None

    if args.handle:
        numbers = args.handle
        person = ', '.join(numbers)
    if args.name:
        person = args.name
        contacts = get_contacts(config)
        if person.lower() not in contacts.keys():
            logger.error(f"{person} not known. Please edit your contacts list.")
            argument_parser.print_help()
            exit(1)

    config.set(CONTROL, 'Person', person)

    if args.output_directory:
        config.set(CONTROL, 'copy directory', args.output_directory)
    if args.no_copy:
        config.set(CONTROL, 'copy', 'False')
    if args.output_type:
        config.set(CONTROL, 'output type', args.output_type)
    if args.force:
        config.set(CONTROL, 'force copy', 'True')
    if args.no_attachments:
        config.set(CONTROL, 'skip attachments', 'True')
    if args.inline:
        config.set(DISPLAY, 'inline attachments', 'True')

    start_date = None
    end_date = None
    if args.start_time:
        try:
            start_date = datetime.strptime(args.start_time, '%Y-%m-%d %H:%M:%S')
        except ValueError as exp:
            argument_parser.print_help(sys.stderr)
            print(f"\n **Start time not correct: {exp}", file=sys.stderr)
            exit(1)
        config.set(CONTROL, 'start time', str(start_date))
    if args.end_time:
        try:
            end_date = datetime.strptime(args.end_time, '%Y-%m-%d %H:%M:%S')
        except ValueError as exp:
            argument_parser.print_help(sys.stderr)
            print(f"\n** End time not correct: {exp}", file=sys.stderr)
            exit(1)
        config.set(CONTROL, 'end time', str(end_date))
    if start_date and end_date and start_date >= end_date:
        argument_parser.print_help(sys.stderr)
        print(f"\n **Start date ({start_date}) must be before end date ({end_date})", file=sys.stderr)
        exit(1)

    out = sys.stdout

    copy_directory = config[CONTROL].get('copy directory', fallback="/tmp")
    attachment_directory = f"{copy_directory}/{person}_attachments"
    config[CONTROL]['attachment directory'] = attachment_directory

    filename = f'{person}.html'
    out = open(f"{copy_directory}/{filename}", 'w')
    try:
        os.mkdir(attachment_directory)
    except FileExistsError:
        pass

    database = DB(args.database, config=config)
    message_list = database.Messages(person, contacts[person.lower()])

    output_type = config[CONTROL].get('output type', fallback='html')
    if output_type == 'text':
        database.TextOutput(args.me, person, message_list, output_file=out).print()
    else:
        database.HTMLOutput(args.me, person, message_list, output_file=out)

    database.disconnect()


if __name__ == '__main__':
    run()



from datetime import datetime


class Chat:
    """ Class for holding information about a chat """

    def __init__(self, database, rowid: str, chat_identifier: str, chat_name: str,
                 last_message_date=None) -> None:
        """
            Parameters
            ----------
            database : imessagedb.DB
                An instance of a connected database

            rowid, chat_identifier, chat_name : str
                The parameters are the fields in the database

            last_message_date : date
                The date the last message in this chat was sent"""

        self._database = database
        self._rowid = rowid
        self._chat_identifier = chat_identifier
        self._chat_name = chat_name
        self._last_message_date = last_message_date
        self._participants = []

    def __repr__(self) -> str:
        return f'{self.rowid}: id => {self.chat_identifier} name => "{self.chat_name}" ' \
               f'last_message => {self.last_message_date}, participants => "{self.participants}"'

    @property
    def rowid(self) -> str:
        return self._rowid

    @property
    def chat_identifier(self) -> str:
        return self._chat_identifier

    @property
    def chat_name(self) -> str:
        return self._chat_name

    @property
    def last_message_date(self) -> datetime:
        return self._last_message_date

    @last_message_date.setter
    def last_message_date(self, date: datetime):
        self._last_message_date = date

    @property
    def participants(self) -> str:
        """ Returns the participants in the chat """
        strings = []
        for handle_id in self._participants:
            if handle_id in self._database.handles.handles:
                strings.append(f'{self._database.handles.handles[handle_id].number} ({handle_id})')
            else:
                strings.append(f'{handle_id}')
        return ', '.join(strings)

    def add_participant(self, participant: str):
        """ Add a participant to the chat """
        if participant not in self._participants:
            self._participants.append(participant)



from datetime import datetime
import re
from alive_progress import alive_bar


def _generate_thread_row(message):
    if message.is_from_me:
        style = 'me'
    else:
        style = 'them'

    text = message.text
    # row_string = f'    <tr><td class="blank></td><td class="reply_text_{style}" > {text} </td ></tr>\n'
    row_string = f'{" ":16s}<tr>\n' \
                 f'{" ":18s}<td class="reply_text_thread">\n' \
                 f'{" ":20s}<a href="#{message.rowid}">\n' \
                 f'{" ":22s}<button class="reply_text_{style}"> {text}</button>\n' \
                 f'{" ":20s}</a>\n' \
                 f'{" ":18s}</td>\n' \
                 f'{" ":16s}</tr>\n'
    return row_string


def _generate_thread_table(message_list, style):
    table_string = f'{" ":14s}<table class="thread_table_{style}">\n'
    for message in message_list:
        table_string = f'{table_string}{_generate_thread_row(message)}'
    table_string = f'{table_string}' \
                   f'{" ":14s}</table>\n' \
                   f'{" ":14s}<p>\n'
    return table_string


def _replace_url_to_link(value):
    """ From https://gist.github.com/guillaumepiot/4539986 """

    # Replace url to link
    urls = re.compile(r"((https?):((//)|(\\\\))+[\w\d:#@%/;$()~_?\+-=\\\.&]*)", re.MULTILINE | re.UNICODE)
    value = urls.sub(r'<a href="\1" target="_blank">\1</a>', value)
    # Replace email to mailto
    urls = re.compile(r"([\w\-\.]+@(\w[\w\-]+\.)+[\w\-]+)", re.MULTILINE | re.UNICODE)
    value = urls.sub(r'<a href="mailto:\1">\1</a>', value)
    return value


class HTMLOutput:
    """ Creates an HTML file (or string) from a Messages list

    ...

    There are a number of options in the configuration file that affect how the HTML is created.
    In the CONTROL section, the following impact the output:

    copy = True :
                If the attachments are not copied, they are not available on the HTML output.
                There are two reasons for that. 1) The web browser does not have access to the directory
                that the attachments are stored, so it cannot display them 2) Some of the attachments need
                to be converted in order to be viewed in the browser, and need a place to live.

    skip attachments = False :
                If true, attachments will not be available in the HTML output

    In the DISPLAY section, the following impact the output:

    inline attachments = False :
                If inline attachments are true, then the images will be part of the HTML. If
                it is false, then there will be a popup when you hover over the attachment text that will show
                the attachment on demand. This will make the load much faster and the output far less cluttered.

    me = Me :
                The name to put for your part of the conversation. It defaults to 'Me'.

    me html background color = AliceBlue
    them html background color = Lavender :
                The background color of the text messages for you and for the other person.
                The options for colors can be found here: https://www.w3schools.com/cssref/css_colors.php


    thread background = HoneyDew
    me thread background = AliceBlue
    them thread background = Lavender :
                The background color of the thread in replies

    me html name color = Blue
    them html name color = Purple :
                The color of the name

    """

    def __init__(self, database, me: str, person: str, message_list,
                 inline=False, output_file=None) -> None:
        """
            Parameters
            ----------
            database : imessagedb.DB
                An instance of a connected database

            me : str
                Your display name

            person : str
                The name of the person in the conversation

            message_list: imessagedb.DB.Messages
                 The messages

            inline : bool
                Display attachments inline or not

            output_filename : str
                The name of the output file"""

        self._database = database
        self._me = me
        self._person = person
        self._message_list = message_list
        self._attachment_list = self._database.attachment_list
        self._inline = inline
        self._output_file = output_file

        self._day = 'UNK'
        self._html_array = []
        self._print_and_save(self._generate_head(), self._html_array)
        self._print_and_save("<body>\n", self._html_array)

        start_time = self._database.control.get('start time', fallback=None)
        end_time = self._database.control.get('end time', fallback=None)
        date_string = ""
        if start_time:
            date_string = f"from {start_time} "
        if end_time:
            date_string = f"{date_string} until {end_time}"

        self._print_and_save(f"Exchanged {len(self._message_list):,} messages with {self._person} {date_string}<p>\n",
                             self._html_array)
        self._print_and_save(f'{" ":2s}<div class="picboxframe"  id="picbox"> <img src="" /> </div>\n',
                             self._html_array)

        self._html_array.append(self._generate_table(self._message_list))
        self._print_and_save('</body>\n</html>\n',
                             self._html_array)

    def __repr__(self) -> str:
        return ''.join(self._html_array)

    def save(self) -> None:
        """ Write the output to the output file"""
        print('\n'.join(self._html_array), file=self._output_file)
        return

    def print(self) -> None:
        """ Print the output to stdout """
        print('\n'.join(self._html_array))
        return

    def _print_and_save(self, message: str, array: list) -> None:
        """ Save to the output file while it is processing """
        array.append(message)
        if self._output_file:
            print(message, end="", file=self._output_file)

    def _generate_table(self, message_list) -> str:
        table_array = []
        self._print_and_save(f'{" ":2s}<table class="main_table">\n', table_array)

        previous_day = ''

        message_count = 0
        with alive_bar(len(message_list), title="Generating HTML", stats="({rate}, eta: {eta})") as bar:
            for message in message_list:
                message_count = message_count + 1

                today = message.date[:10]
                if today != previous_day:
                    previous_day = today

                    # If it's a new day, end the table, and start a new one
                    self._print_and_save(f'{" ":2s}</table>\n\n{" ":2s}<table class="main_table">\n', table_array)

                    self._day = datetime(int(message.date[0:4]), int(message.date[5:7]), int(message.date[8:10]),
                                         int(message.date[11:13]), int(message.date[14:16]),
                                         int(message.date[17:19])).strftime('%a')
                self._print_and_save(self._generate_row(message), table_array)
                if message.attachments:
                    bar()
                else:
                    bar(skipped=True)

        self._print_and_save(f'{" ":2s}</table>\n', table_array)
        return ''.join(table_array)

    def _generate_row(self, message) -> str:
        # Specify if the message is from me, or the other person
        if message.is_from_me:
            who = self._me
            style = 'me'
        else:
            who = self._person
            style = 'them'

        # Check to see if we want the media box floating or fixed
        floating_option = self._database.config['DISPLAY'].get('popup location', fallback='floating')
        floating = floating_option == 'floating'

        # If this message is part of a thread, then show the messages in the thread before it
        thread_table = ""
        if message.thread_originator_guid:
            if message.thread_originator_guid in self._message_list.guids:
                original_message = self._message_list.guids[message.thread_originator_guid]
                thread_list = original_message.thread
                thread_list[original_message.rowid] = original_message
                print_thread = []
                # sort the threads by the date sent
                for i in sorted(thread_list.values(), key=lambda x: x.date):
                    if i == message: # stop at the current message
                        break
                    print_thread.append(i)
                thread_table = _generate_thread_table(print_thread, style)

        # Generate the attatchment string
        attachments_string = ""
        if message.attachments:
            for attachment_key in message.attachments:
                if attachment_key not in self._attachment_list.attachment_list:
                    attachments_string = f'{attachments_string} <span class="missing"> Attachment missing </span> '
                    continue

                attachment = self._attachment_list.attachment_list[attachment_key]
                attachment_string = ""
                if attachment.skip:
                    continue
                if attachment.missing:
                    attachment_string = f'{attachments_string} <span class="missing"> Attachment missing </span> '
                    continue
                if attachment.copy:
                    if attachment.conversion_type == 'HEIC':
                        attachment.convert_heic_image(attachment.original_path, attachment.destination_path)
                    elif attachment.conversion_type == 'Audio' or attachment.conversion_type == 'Video':
                        attachment.convert_audio_video(attachment.original_path, attachment.destination_path)
                    else:
                        attachment.copy_attachment()

                if floating:
                    box_name = f'PopUp{attachment.rowid}'
                    image_box = f'<div class="imageBox" id="PopUp{attachment.rowid}">  <img src="" /> </div>'
                else:
                    box_name = 'picbox'
                    image_box = ''

                if attachment.popup_type == 'Picture':
                    if self._inline:
                        attachment_string = f'<p><a href="{attachment.html_path}" target="_blank">' \
                                            f'<img src="{attachment.html_path}" target="_blank"/><p>' \
                                            f' {attachment.html_path} </a>\n'
                    else:
                        attachment_string = f'''<a href="{attachment.html_path}" target="_blank"
            onMouseOver="ShowPicture('{box_name}',1,'{attachment.html_path}')" 
            onMouseOut="ShowPicture('{box_name}',0)"> {attachment.html_path} </a>
'''
                elif attachment.popup_type == 'Audio':
                    # Not going to do popups for audio, just inline
                    attachment_string = f'<p><audio controls>  <source src="{attachment.html_path}" ' \
                                        f'type="audio/mp3"></audio> <a href="{attachment.html_path}" ' \
                                        f'target="_blank"> {attachment.html_path} </a>\n'
                elif attachment.popup_type == 'Video':
                    if self._inline:
                        attachment_string = f'<p><video controls>  <source src="{attachment.html_path}" ' \
                                            f' type="video/mp4"></video> <p><a href="{attachment.html_path}"' \
                                            f' target="_blank"> {attachment.html_path} </a>\n'
                    else:
                        attachment_string = f'''<a href="{attachment.html_path}" target="_blank"
            onMouseOver="ShowMovie('{box_name}', 1, '{attachment.html_path}')"> {attachment.html_path} </a>
'''

                else:
                    attachment_string = f'<a href="{attachment.html_path}" target="_blank"> ' \
                                        f'{attachment.html_path} </a>\n'

                attachments_string = f'{attachments_string} <p> {attachment_string} {image_box} '

        # Structure of the text row. The first three columns are normal, the fourth column is complex
        #   <tr>
        #     <td> date
        #     <td> who
        #     <td>
        #       <table>
        #         <tr hidden>
        #           <td> Edited Row
        #         </tr>
        #         <tr>
        #           <td> Text of message (edited if required)
        #           <td hidden> extra text
        #           <td> info button (if configured)
        #         <tr>

        text = _replace_url_to_link(f'{message.text} {attachments_string}')

        # Check for edits on the text. If there are edits, then set up the html to allow for that. The row
        #   doesn't exist if there is no edits

        edited_string = ""
        edit_table = ""
        text_cell_edit_row = ""

        if len(message.edits) > 0:
            edit_table = f'{" ":14s}<div class="edits_{style}">\n'
            for i in range(0, len(message.edits)):
                edit_table = f'{edit_table}{" ":16s}"{message.edits[i]} <p>\n'
            edit_table = f'{edit_table}{" ":14s}</div>\n'
            edited_string = f'<sub><button class="edited_button"' \
                            f' onclick="ToggleDisplay(\'{message.rowid}editTable\')"> Edited </button></sub>'
            text_cell_edit_row = f'{" ":10s}<tr id={message.rowid}editTable class="edits">\n' \
                                 f'{" ":12s}<td>\n' \
                                 f'{edit_table} ' \
                                 f'{" ":12s}</td>\n ' \
                                 f'{" ":10s}</tr>\n'

        # Check for if we want additional data

        info_text = ""
        info_button = ""
        if self._database.control.getboolean('additional details', fallback=False):
            info_text = f'{" ":12s}<td class="infocell" id={message.rowid}info>\n ' \
                        f'{" ":14s}<table>\n' \
                        f'{" ":16s}<tr>\n' \
                        f'{" ":18s}<td> ChatID: {message.chat_id} </td>\n' \
                        f'{" ":16s}</tr>\n' \
                        f'{" ":14s}</table>\n' \
                        f'{" ":12s}</td>\n'
            info_button = f'{" ":12s}<td class="button-wrapper"> <button class="text_{style}" ' \
                          f'onclick="ToggleDisplay(\'{message.rowid}info\')"> ℹ️ </button> </td>\n'

        # Put together the row cells

        date_cell = f'{" ":6s}<td class="date"> {self._day} {message.date} </td>\n'
        name_cell = f'{" ":6s}<td class="name_{style}"> {who}: </td>\n'

        text_cell = f'{" ":6s}<td>\n ' \
                    f'{" ":8s}<table>\n' \
                    f'{text_cell_edit_row}' \
                    f'{" ":10s}<tr>\n' \
                    f'{" ":12s}<td class="text_{style}">\n' \
                    f'{thread_table}' \
                    f'{" ":14s}{text} {edited_string}\n' \
                    f'{" ":12s}</td>\n' \
                    f'{info_text}' \
                    f'{info_button}' \
                    f'{" ":10s}</tr>\n' \
                    f'{" ":8s}</table>\n' \
                    f'{" ":6s}</td>\n'

        row_string = f'{" ":4s}<tr id={message.rowid}>\n' \
                     f'{date_cell}' \
                     f'{name_cell}' \
                     f'{text_cell}' \
                     f'{" ":4s}</tr>\n'

        return row_string

    def _generate_head(self) -> str:
        popup = self._database.config['DISPLAY'].get('popup location', fallback='upper right')
        if popup == 'upper right':
            popup_location = 'right'
        else:
            popup_location = 'left'

        me_html_background_color = self._database.config['DISPLAY'].get('me html background_color',
                                                                        fallback='AliceBlue')
        them_html_background_color = self._database.config['DISPLAY'].get('them html background color',
                                                                          fallback='Lavender')
        thread_background_color = self._database.config['DISPLAY'].get('thread background',
                                                                       fallback='HoneyDew')
        me_thread_background_color = self._database.config['DISPLAY'].get('me thread background',
                                                                          fallback='AliceBlue')
        them_thread_background_color = self._database.config['DISPLAY'].get('them thread background',
                                                                            fallback='Lavender')
        me_html_name_color = self._database.config['DISPLAY'].get('me html name color',
                                                                  fallback='Blue')
        them_html_name_color = self._database.config['DISPLAY'].get('them html name color',
                                                                    fallback='Purple')

        css = '''    <style>
table {''' + f'''
    width: 100%;
    table-layout: auto;
''' + ''' }

table.main_table {''' + f'''
    border-bottom: 3px solid black;
    border-spacing: 8px;
''' + ''' }

table.thread_table_me {''' + f'''
    width: 50%;
    margin-right: 0px;
    margin-left: auto;
    background: {thread_background_color};
    padding: 5px;
    border-radius: 30px;
''' + ''' }

table.thread_table_them {''' + f'''
    width: 50%;
    margin-right: auto;
    margin-left: 0px;
    background: {thread_background_color};
    padding: 5px;
    border-radius: 30px;
''' + ''' }

td {''' + f'''
    padding: 0px;
    margin: 0;
    line-height: 1;
''' + ''' }

td.date {''' + f'''
    text-align: left;
    width: 150px;
    vertical-align: text-middle;
    font-size: 80%;
''' + ''' }

td.name_me {''' + f'''
    text-align: right;
    font-weight: bold;
    color: {me_html_name_color};
    width: 50px;
    padding-right: 5px;
    vertical-align: text-middle;
    font-size: 80%;
    
''' + ''' }

td.name_them {''' + f'''
    text-align: right;
    font-weight: bold;
    color: {them_html_name_color};
    width: 50px;
    padding-right: 5px;
    vertical-align: text-middle;
    font-size: 80%;
''' + ''' }

td.text_me {''' + f'''
    text-align: right;
    word-wrap: break-word;
    border-radius: 30px;
    padding: 15px;
    border-spacing: 40px;
    background: {me_html_background_color};  
''' + ''' }

td.text_them {''' + f'''
    text-align: left;
    word-wrap: break-word;
    border-radius: 30px;
    padding: 15px;
    border-spacing: 40px;
    background: {them_html_background_color};
''' + ''' }

.edits_me {''' + f'''
    display: none;
    font-size: 70%;
    font-style: italics;
    text-align: right;
    border-radius: 30px;
''' + ''' }  

.edits_them {''' + f'''
    display: none;
    font-size: 70%;
    font-style: italics;
    text-align: left;
    border-radius: 30px;
''' + ''' }    

.infocell {''' + f'''
    margin-right: 0px;
    margin-left: auto;
    width: 20%;
    text-align: right;
    font-size: 70%;
    display: none;
''' + ''' }

button.edited_button {''' + f'''
    font-size: 50%;
    border-radius: 30px;
    font-size: 50%;
    padding-left: 0px;
    padding-right: 0px;
    border-spacing: 0px;
    border-bottom: 0px;
    margin-right: 0px;
    margin-left: 0px;
    text-align: center;
    border: 0px;
    color: blue;
    font-style: italic;
    background-color: transparent;
''' + ''' }

td.button-wrapper {''' + f'''
    margin-right: 0px;
    margin-left: 4px;
    padding-left: 0px;
    padding-right: 0px;
    border-spacing: 0px;
    text-align: center;
    width:0.1%;
    background-color: transparent; 
''' + ''' }

button.text_me {''' + f'''
    background: {me_html_background_color};
    border-radius: 30px;
    font-size: 50%;
    padding-left: 0px;
    padding-right: 0px;
    border-spacing: 0px;
    border-bottom: 0px;
    margin-right: 0px;
    margin-left: 0px;
    text-align: center;
    border: 0px;
''' + ''' }

button.text_them {''' + f'''
    background: {them_html_background_color};
    border-radius: 30px;
    font-size: 50%;
    padding-left: 0px;
    padding-right: 0px;
    border-spacing: 0px;
    border-bottom: 0px;
    margin-right: 0px;
    margin-left: 0px;
    text-align: center;
    border: 0px;
''' + ''' }

reply_text_thread {''' + f'''
    border: 2px solid;
    background: {thread_background_color};
    border-radius: 6px;
    border-radius: 50px;
    font-size: 60%
''' + ''' }

.reply_text_me {''' + f'''
    border: 2px solid;
    background: {me_html_background_color};
    border-radius: 6px;
    border-radius: 50px;
    font-size: 60%
''' + ''' }

.reply_text_them {''' + f'''
    border: 2px solid;
    background: {them_html_background_color};
    border-radius: 6px;
    border-radius: 50px;
    font-size: 60%
''' + ''' }

td.blank {''' + f'''
    border: none;
    width: 50%
''' + ''' }

.missing {''' + f'''
    color: red;
''' + ''' }

.badjoin {''' + f'''
    color: red;
''' + ''' }

.imageBox {''' + f'''
    position: absolute;
    visibility: hidden;
    height: 200;
    border: solid 1px #CCC;
    padding: 5px;
''' + ''' }

img {''' + f'''
    height: 250px;
    width: auto;
''' + ''' }

.picboxframe {''' + f'''
    position: fixed;
    top: 2%;
    ''' + f'{popup_location}: 2%;' \
          '''
    background: Blue;
    transition: all .5s ease;

''' + ''' }

    </style>'''
        script = '''  <script>
    
    function ToggleDisplay(id) {
      if (document.getElementById(id).style.display == "none") {
          document.getElementById(id).style.display = "inline";
      }
      else {
          document.getElementById(id).style.display = "none";
      }
    }
    
    function ShowPicture(id,show, img) {
      if (show=="1") {
        document.getElementById(id).style.visibility = "visible"
        document.getElementById(id).childNodes[1].src = img;
      }
      else if (show=="0") {
        document.getElementById(id).style.visibility = "hidden"
      }
    }

    function ShowMovie(id, show, movie) {
        var elem = document.getElementById(id);
''' \
                 f'        var htmlstring = "<video controls onMouseOut=\'ShowMovie(\\\"\" + id + "\\\",0)\'> ' \
                 f'<source src=\'" + movie + "\'> </video>";' \
                 '''
        if (show == "1") {
          {
            elem.style.visibility = "visible";
            elem.innerHTML = htmlstring;
          }
        } else if (show == "0") {
          {
            elem.style.visibility = "hidden";
            elem.innerHTML =  " "
          }
        }
      }
  </script>'''

        head_string = '''<!DOCTYPE html>
<html lang="en-US">
  <head>
    <meta http-equiv="Content-Type" content="text/html; charset=utf-8" />
    ''' \
                      f'    <title> {self._person} </title>\n{css}\n{script}\n</head>\n'
        return head_string




class Handle:
    """ Class for holding information about a Handle """

    def __init__(self, database, rowid: str, number: str, service: str) -> None:
        """
            Parameters
            ----------
            database : imessagedb.DB
                An instance of a connected database

            rowid, number, service : str
                The parameters are the fields in the database"""

        self._database = database
        self._rowid = rowid
        self._number = number
        self._service = service
        self._name = "Unknown"

    def __repr__(self) -> str:
        return f'{self._rowid}: ID => {self._number}, Service => {self._service}, Name => {self._name}'

    @property
    def rowid(self) -> str:
        return self._rowid

    @property
    def number(self) -> str:
        return self._number

    @property
    def service(self) -> str:
        return self._service

    @property
    def name(self) -> str:
        return self._name




from datetime import datetime
from alive_progress import alive_bar
from imessagedb.message import Message


mac_epoch_start = int(datetime(2001, 1, 1, 0, 0, 0).strftime('%s'))


def _convert_to_database_date(date_string):
    date_ = datetime.strptime(date_string, '%Y-%m-%d %H:%M:%S')
    epoch_date = int(date_.strftime('%s'))
    diff = epoch_date - mac_epoch_start
    return diff * 1000000000


def _convert_from_database_date(date_value):
    date_ = date_value / 1000000000
    epoch_date = date_ + mac_epoch_start
    return datetime.fromtimestamp(epoch_date)


class Messages:
    """ All messages in a conversation or conversations with a particular person """
    def __init__(self, database, person: str, numbers: list) -> None:
        """
                Parameters
                ----------
                database : imessagedb.DB
                    An instance of a connected database

                person : str
                    The name of the person in the conversation

                numbers: list
                    A list of numbers associated with the person, as represented in the handle data table

                """
        self._database = database
        self._person = person
        self._numbers = numbers
        self._guids = {}
        self._message_list = {}

        numbers_string = "','".join(self._numbers)
        time_rules = []
        time_where_clause = ""
        start_time = self._database.control.get('start time', fallback=None)
        end_time = self._database.control.get('end time', fallback=None)
        if start_time:
            database_start_date = _convert_to_database_date(start_time)
            time_rules.append(f"message.date >= {database_start_date}")
        if end_time:
            database_end_date = _convert_to_database_date(end_time)
            time_rules.append(f"message.date <= {database_end_date}")
        if len(time_rules) > 0:
            time_where_clause = f" and {' AND '.join(time_rules)}"

        where_clause = "rowid in (" \
                       " select message_id from chat_message_join where chat_id in (" \
                       "  select chat_id from chat_handle_join where handle_id in (" \
                       f"   select rowid from handle where id in ('{numbers_string}')" \
                       "  )" \
                       " )" \
                       f") {time_where_clause}"
        select_string = "select message.rowid, guid, " \
                        "datetime(message.date/1000000000 + strftime('%s', '2001-01-01'),'unixepoch','localtime'), " \
                        "message.is_from_me, message.handle_id, " \
                        " message.attributedBody, message.message_summary_info, message.text, " \
                        "reply_to_guid, thread_originator_guid, thread_originator_part, cmj.chat_id  " \
                        "from message, chat_message_join cmj " \
                        f"where message.rowid = cmj.message_id and {where_clause} " \
                        "order by message.date asc"
        row_count_string = f"select count (*) from message where {where_clause}"

        self._database.connection.execute(row_count_string)
        (row_count_total) = self._database.connection.fetchone()
        row_count_total = row_count_total[0]

        self._database.connection.execute(select_string)

        i = self._database.connection.fetchone()

        with alive_bar(row_count_total, title="Getting Messages", stats="({rate}, eta: {eta})") as bar:
            message_count = 0
            while i:
                (rowid, guid, date, is_from_me, handle_id, attributed_body, message_summary_info, text,
                 reply_to_guid, thread_originator_guid, thread_originator_part, chat_id) = i
                message_count = message_count + 1

                skip_attachment = self._database.control.getboolean('skip attachments', fallback=False)
                attachment_list = None
                if not skip_attachment:
                    if rowid in self._database.attachment_list.message_join:
                        attachment_list = self._database.attachment_list.message_join[rowid]

                skipped = True

                new_message = Message(self._database, rowid, guid, date, is_from_me, handle_id, attributed_body,
                                      message_summary_info, text, reply_to_guid, thread_originator_guid,
                                      thread_originator_part, chat_id, attachment_list)
                self._guids[guid] = new_message
                self._message_list[rowid] = new_message

                # Manage the thread
                if thread_originator_guid:
                    self._guids[thread_originator_guid].thread[rowid] = new_message

                bar(skipped=skipped)
                i = self._database.connection.fetchone()

        self._sorted_message_list = sorted(self._message_list.values(), key=lambda x: x.date)

    @property
    def message_list(self) -> list:
        """ Returns a list of messages sorted by the date of the message"""
        return self._sorted_message_list

    @property
    def guids(self) -> dict:
        return self._guids

    def __iter__(self):
        return self._sorted_message_list.__iter__()

    def __len__(self) -> int:
        return len(self._sorted_message_list)



import sqlite3
import os
import configparser

import imessagedb
from imessagedb.attachments import Attachments
from imessagedb.messages import Messages
from imessagedb.message import Message
from imessagedb.chats import Chats
from imessagedb.handles import Handles
from imessagedb.html import HTMLOutput
from imessagedb.text import TextOutput


class DB:
    """
    A class to connect to an iMessage database

    """
    def __init__(self, database_name=None, config=None):
        """
        Parameters
        ----------
        database_name : str
            The database that it connects to (the default is to use the default database in the caller's home directory)

        config : ConfigParser
            Configuration information. If none is provided, we will create a default.
        """

        if database_name is None:
            database_name = f"{os.environ['HOME']}/Library/Messages/chat.db"
        self._database_name = database_name
        self._configuration = config

        if self._configuration is None:
            self._configuration = configparser.ConfigParser()
            self._configuration.read_string(imessagedb.DEFAULT_CONFIGURATION)

        self._control = self._configuration['CONTROL']

        self._chat_connection = sqlite3.connect(database_name)
        self._cursor = self._chat_connection.cursor()

        # Preload some of the data
        self._handles = Handles(self)
        self._chats = Chats(self)
        self._attachment_list = Attachments(self)
        return

    def Messages(self, person: str, numbers: list) -> Messages:
        """A wrapper to create a Messages class
        """
        return Messages(self, person, numbers)

    def HTMLOutput(self, me: str, person: str, message_list: Messages, inline=False, output_file=None) -> HTMLOutput:
        """A wrapper to create an HTMLOutput class
        """
        return HTMLOutput(self, me, person, message_list, inline, output_file)

    def TextOutput(self, me: str, person: str, message_list: Messages, output_file=None) -> TextOutput:
        """A wrapper to create a TextOutput class
        """
        return TextOutput(self, me, person, message_list, output_file)

    def disconnect(self) -> None:
        """Disconnects from the database

        """
        self._chat_connection.close()
        return

    @property
    def connection(self) -> sqlite3.Cursor:
        """Returns a connection to query the database
        """
        return self._cursor

    @property
    def handles(self) -> Handles:
        """Returns an imessagedb.Handles class with all the handles
        """
        return self._handles

    @property
    def chats(self) -> Chats:
        """Returns an imessagedb.Chats class with all the chats
        """
        return self._chats

    @property
    def attachment_list(self) -> Attachments:
        """Returns an imessagedb.Attachments class with all the attachments
        """
        return self._attachment_list

    @property
    def config(self) -> configparser.ConfigParser:
        """Returns the configuration object
        """
        return self._configuration

    @property
    def control(self):
        """Returns a shortcut to the CONTROL section of the configuration
                """
        return self._control
